#include <v_lib.h>

/*
 * I hate C++, but this language seems to be good for debugging ... :)
 */

struct vector_c
{
	int		base, len;
	vec_t	stub;
	vec_p	data, idx;
	vboolean verbose;
	char	caller[64];
	vector_c(int base,int len)
	{
		this->base = base;
		this->len = len;
		data = (vec_p)memAlloc(len * sizeof(vec_t));
		idx = data - base;
		verbose = false;
		strcpy(caller,"vector_c::operator[]");
	}
	~vector_c()
	{
		memFree(data);
	}
	vec_t &operator[](int elem)
	{
		if(elem < base || (elem-base)>=len)
		{
			conlPrintf(LOG_ERROR,
				"%s: wrong col (lower %d, upper %d, req %d).\n",
				caller,base,base+len-1,elem);
			stub = 0;
			return stub;
		}
		if(verbose)
			conlPrintf(LOG_INFO,
				"%s: col lower %d, upper %d, req %d.\n",
				caller,base,base + len - 1,elem);
		return idx[elem];
	}
};

struct matrix_c
{
	int		row_base, num_rows;
	int		col_base, num_cols;
	vector_c **rows;
	matrix_c(int row_base,int num_rows,int col_base,int num_cols)
	{
		this->row_base = row_base;
		this->num_rows = num_rows;
		this->col_base = col_base;
		this->num_cols = num_cols;
		rows = (vector_c**)memAlloc((num_rows + 1) * sizeof(vector_c*));
		for(int i=0;i <= num_rows;i++)
		{
			rows[i] = new vector_c(col_base,num_cols);
			strcpy(rows[i]->caller,"matrix_c::operator[][]");
		}
		rows[num_rows]->verbose = true;
	}
	~matrix_c(void)
	{
		for(int i=0;i <= num_rows;i++)
			delete rows[i];
		memFree(rows);
	}
	vector_c &operator[](int row_index)
	{
		if(row_index < row_base || (row_index - row_base) >= num_rows)
		{
			conlPrintf(LOG_ERROR,
				"matrix_c::operator[]: wrong row (lower %d, upper %d, req %d).\n",
				row_base,row_base + num_rows - 1,row_index);
			return *(rows[num_rows]);
		}
		return *(rows[row_index - row_base]);
	}
};

#if 0
/* debug */

void mmid(vec_t y[], vec_t dydx[], int nvar, vec_t xs, vec_t htot,
	int nstep,vec_t yout[], void (*derivs)(vec_t, vec_t[], vec_t[]))
{
	/*
	 * Modified midpoint step. At xs, input the dependent variable
	 * vector y[1..nvar] and its derivative vector dydx[1..nvar].
	 * Also input is htot, the total step to be made, and nstep,
	 * the number of substeps to be used.
	 * The output is returned as yout[1..nvar], which need not be a
	 * distinct array from y; if it is distinct, however, then y and
	 * dydx are returned undamaged.
	 */
	int n, i;
	vec_t x, swap, h2, h, *ym, *yn;
	ym = nr_vector(1,nvar);
	yn = nr_vector(1,nvar);
	h = htot / nstep; /* Stepsize this trip. */
	for(i = 1;i <= nvar;i++)
	{
		ym[i] = y[i];
		yn[i] = y[i] + h * dydx[i]; /* First step. */
	}
	x = xs + h;
	(*derivs)(x,yn,yout); /* Will use yout for temporary storage of derivatives. */
	h2 = 2.0 * h;
	for(n = 2;n <= nstep;n++) /* General step. */
	{
		for(i = 1;i <= nvar;i++)
		{
			swap = ym[i] + h2 * yout[i];
			ym[i] = yn[i];
			yn[i] = swap;
		}
		x += h;
		(*derivs)(x,yn,yout);
	}
	for(i = 1;i <= nvar;i++) /* Last step. */
		yout[i] = 0.5*(ym[i] + yn[i] + h*yout[i]);
	nr_free_vector(yn,1);
	nr_free_vector(ym,1);
}

#define KMAXX 8 /* Maximum row number used in the extrapolation. */
#define IMAXX (KMAXX+1)
#define SAFE1 0.25 /* Safety factors. */
#define SAFE2 0.7
#define REDMAX 1.0e-5 /* Maximum factor for stepsize reduction. */
#define REDMIN 0.7 /* Minimum factor for stepsize reduction. */
#define TINY 1.0e-30 /* Prevents division by zero. */
#define SCALMX 0.1 /* 1/SCALMX is the maximum factor by which a stepsize can be increased. */

vec_t **d, *x;
/*Pointers to matrix and vector used by pzextr or rzextr. */

void
pzextr(int iest, vec_t xest, vec_t yest[], vec_t yz[], vec_t dy[], int nv)
/*
Use polynomial extrapolation to evaluate nv functions at x = 0 by fitting a polynomial to a
sequence of estimates with progressively smaller values x = xest, and corresponding function
vectors yest[1..nv]. This call is number iest in the sequence of calls. Extrapolated function
values are output as yz[1..nv], and their estimated error is output as dy[1..nv].
*/
{
	int k1, j;
	vec_t q, f2, f1, delta, *c;
	c = nr_vector(1, nv);
	x[iest] = xest;				/* Save current independent variable. */
	for (j = 1; j <= nv; j++)
		dy[j] = yz[j] = yest[j];
	if (iest == 1) {			/* Store first estimate infirst column. */
		for (j = 1; j <= nv; j++)
			d[j][1] = yest[j];
	} else {
		for (j = 1; j <= nv; j++)
			c[j] = yest[j];
		for (k1 = 1; k1 < iest; k1++) {
			delta = 1.0 / (x[iest - k1] - xest);
			f1 = xest * delta;
			f2 = x[iest - k1] * delta;
			for (j = 1; j <= nv; j++) {	/* Propagate tableau 1 diagonal more. */
				q = d[j][k1];
				d[j][k1] = dy[j];
				delta = c[j] - q;
				dy[j] = f1 * delta;
				c[j] = f2 * delta;
				yz[j] += dy[j];
			}
		}
		for (j = 1; j <= nv; j++)
			d[j][iest] = dy[j];
	}
	nr_free_vector(c, 1);
}

void
rzextr(int iest, vec_t xest, vec_t yest[], vec_t yz[], vec_t dy[], int nv)
/* Exact substitute for pzextr, but uses diagonal rational function extrapolation instead of poly-
nomial extrapolation. */
{
	int k, j;
	vec_t yy, v, ddy = 0, c, b1, b, *fx;
	fx = nr_vector(1, iest);
	x[iest] = xest;				/* Save current independent variable. */
	if (iest == 1)
		for (j = 1; j <= nv; j++) {
			yz[j] = yest[j];
			d[j][1] = yest[j];
			dy[j] = yest[j];
	} else {
		for (k = 1; k < iest; k++)
			fx[k + 1] = x[iest - k] / xest;
		for (j = 1; j <= nv; j++) {	/* Evaluate next diagonal in tableau. */
			v = d[j][1];
			d[j][1] = yy = c = yest[j];
			for (k = 2; k <= iest; k++) {
				b1 = fx[k] * v;
				b = b1 - c;
				if (b) {
					b = (c - v) / b;
					ddy = c * b;
					c = b1 * b;
				} else			/* Care needed to avoid division by 0. */
					ddy = v;
				if (k != iest)
					v = d[j][k];
				d[j][k] = ddy;
				yy += ddy;
			}
			dy[j] = ddy;
			yz[j] = yy;
		}
	}
	nr_free_vector(fx, 1);
}

#define SQR(dest,src) {vec_t tmp; tmp = src; dest = tmp * tmp; }
#define FMAX(dest,v1,v2) {vec_t t1, t2; t1 = v1; t2 = v2; dest = t1 > t2 ? t1 : t2; }
#define FMIN(dest,v1,v2) {vec_t t1, t2; t1 = v1; t2 = v2; dest = t1 > t2 ? t2 : t1; }

void
bsstep(vec_t y[], vec_t dydx[], int nv, vec_t *xx, vec_t htry, vec_t eps,
	   vec_t yscal[], vec_t *hdid, vec_t *hnext,
	   void (*derivs) (vec_t, vec_t[], vec_t[]))
/*
Bulirsch-Stoer step with monitoring of local truncation error to ensure accuracy and adjust
stepsize. Input are the dependent variable vector y[1..nv] and its derivative dydx[1..nv]
at the starting value of the independent variable x. Also input are the stepsize to be attempted
htry, the required accuracy eps, and the vector yscal[1..nv] against which the error is
scaled. On output, y and x are replaced by their new values, hdid is the stepsize that was
actually accomplished, and hnext is the estimated next stepsize. derivs is the user-supplied
routine that computes the right-hand side derivatives. Be sure to set htry on successive steps
to the value of hnext returned from the previous step, as is the case if the routine is called
by odeint.
*/
{
	int i, iq, k, kk, km = 0;
	static int first = 1, kmax = KMAXX, kopt;
	static vec_t epsold = -1.0, xnew;
	vec_t eps1, errmax = 0, fact, h, red = 0, scale = 0, work, wrkmin, xest;
	vec_t *err, *yerr, *ysav, *yseq;
	static vec_t a[IMAXX + 1];
	static vec_t alf[KMAXX + 1][KMAXX + 1];
	static int nseq[IMAXX + 1] = { 0, 2, 4, 6, 8, 10, 12, 14, 16, 18 };
	int reduct, exitflag = 0;
	d = nr_matrix(1, nv, 1, KMAXX);
	err = nr_vector(1, KMAXX);
	x = nr_vector(1, KMAXX);
	yerr = nr_vector(1, nv);
	ysav = nr_vector(1, nv);
	yseq = nr_vector(1, nv);
	if (eps != epsold) {		/* A new tolerance, so reinitialize. */
		*hnext = xnew = -1.0e29;	/* "Impossible" values. */
		eps1 = SAFE1 * eps;
		a[1] = nseq[1] + 1;		/* Compute work coeficients A_k. */
		for (k = 1; k <= KMAXX; k++)
			a[k + 1] = a[k] + nseq[k + 1];
		for (iq = 2; iq <= KMAXX; iq++) {	/* Compute alpha(k; q). */
			for (k = 1; k < iq; k++)
				alf[k][iq] = pow(eps1, (a[k + 1] - a[iq + 1]) /
								 ((a[iq + 1] - a[1] + 1.0) * (2 * k + 1)));
		}
		epsold = eps;
		for (kopt = 2; kopt < KMAXX; kopt++)	/* Determine optimal row number for
												   convergence. */
			if (a[kopt + 1] > a[kopt] * alf[kopt - 1][kopt])
				break;
		kmax = kopt;
	}
	h = htry;
	for (i = 1; i <= nv; i++)
		ysav[i] = y[i];			/* Save the starting values. */
	if (*xx != xnew || h != (*hnext)) {	/* A new stepsize or a new integration:
										   re-establish the order window. */
		first = 1;
		kopt = kmax;
	}
	reduct = 0;
	for (;;) {
		for (k = 1; k <= kmax; k++) {	/* Evaluate the sequence of modified
										   midpoint integrations. */
			xnew = (*xx) + h;
			if (xnew == (*xx))
			{
				puts("step size underflow in bsstep");
				return;
			}
			mmid(ysav, dydx, nv, *xx, h, nseq[k], yseq, derivs);
			SQR(xest, h / nseq[k]);	/* Squared, since error series is even. */
			pzextr(k, xest, yseq, y, yerr, nv);	/* Perform extrapolation. */
			if (k != 1) {		/* Compute normalized error estimate
								   eps(k). */
				errmax = TINY;
				for (i = 1; i <= nv; i++)
					FMAX(errmax, errmax, fabs(yerr[i] / yscal[i]));
				errmax /= eps;	/* Scale error relative to tolerance. */
				km = k - 1;
				err[km] = pow(errmax / SAFE1, 1.0 / (2 * km + 1));
			}
			if (k != 1 && (k >= kopt - 1 || first)) {	/* In order window. */
				if (errmax < 1.0) {	/* Converged. */
					exitflag = 1;
					break;
				}
				if (k == kmax || k == kopt + 1) {	/* Check for possible stepsize
													   reduction. */
					red = SAFE2 / err[km];
					break;
				} else if (k == kopt && alf[kopt - 1][kopt] < err[km]) {
					red = 1.0 / err[km];
					break;
				} else if (kopt == kmax && alf[km][kmax - 1] < err[km]) {
					red = alf[km][kmax - 1] * SAFE2 / err[km];
					break;
				} else if (alf[km][kopt] < err[km]) {
					red = alf[km][kopt - 1] / err[km];
					break;
				}
			}
		}
		if (exitflag)
			break;
		FMIN(red, red, REDMIN);	/* Reduce stepsize by at least REDMIN
									   and at most REDMAX. */
		FMAX(red, red, REDMAX);
		h *= red;
		reduct = 1;
	}							/* Try agai n. */
	*xx = xnew;					/* Successful step taken. */
	*hdid = h;
	first = 0;
	wrkmin = 1.0e35;			/* Compute optimal row for convergence
								   and corresponding stepsize. */
	for (kk = 1; kk <= km; kk++) {
		FMAX(fact, err[kk], SCALMX);
		work = fact * a[kk + 1];
		if (work < wrkmin) {
			scale = fact;
			wrkmin = work;
			kopt = kk + 1;
		}
	}
	*hnext = h / scale;
	if (kopt >= k && kopt != kmax && !reduct) {
/* Check for possible order increase, but not if stepsize was just reduced. */
		FMAX(fact, scale / alf[kopt - 1][kopt], SCALMX);
		if (a[kopt + 1] * fact <= wrkmin) {
			*hnext = h / fact;
			kopt++;
		}
	}
	nr_free_vector(ysav, 1);
	nr_free_vector(yerr, 1);
	nr_free_vector(x, 1);
	nr_free_vector(err, 1);
	nr_free_matrix(d, 1);
	nr_free_vector(yseq, 1);
}

#define MAXSTP 10000
#define SIGN(modulus,signum) (((signum) >= 0) ? v_abs(modulus) : -v_abs(modulus))

void odeint2(vec_t ystart[], int nvar, vec_t x1, vec_t x2, vec_t eps, vec_t h1,
	vec_t hmin,void (*derivs)(vec_t, vec_t [], vec_t []))
{
	int nstp, i;
	vec_t x, hnext, hdid, h;
	vec_t *yscal, *y, *dydx;
	yscal = nr_vector(1,nvar);
	y = nr_vector(1,nvar);
	dydx = nr_vector(1,nvar);
	x = x1;
	h = SIGN(h1,x2 - x1);
	for (i=1;i<=nvar;i++)
		y[i] = ystart[i];
	for (nstp=1;nstp<=MAXSTP;nstp++)
	{
		(*derivs)(x,y,dydx);
		for (i=1;i<=nvar;i++)
			yscal[i]=fabs(y[i])+fabs(dydx[i]*h)+TINY;
		if((x+h - x2) * (x+h - x1) > 0.0)
			h = x2 - x;
		bsstep(y,dydx,nvar,&x,h,eps,yscal,&hdid,&hnext,derivs);
		if((x - x2) * (x2 - x1) >= 0.0)
		{
			for (i=1;i<=nvar;i++)
				ystart[i]=y[i];
			nr_free_vector(dydx,1);
			nr_free_vector(y,1);
			nr_free_vector(yscal,1);
			return;
		}
		if(fabs(hnext) <= hmin)
		{
			puts("Step size too small in odeint");
			return;
		}
		h = hnext;
	}
	puts("Too many steps in routine odeint");
}







void derivs(vec_t x,vec_t y[],vec_t dydx[])
{
	/* y = C * exp(x^2 / 2) */
	dydx[0] = x * y[0];
}

#define	MATRIX_CONSTRUCTOR(STORAGE,ROUTINE,BASE_TYPE) \
STORAGE BASE_TYPE \
**ROUTINE(int col_base,int num_cols,int row_base,int num_rows) \
{ \
	int r; \
	BASE_TYPE **idx; \
	printf("data size %u\n",num_cols * num_rows * sizeof(BASE_TYPE)); \
	printf("idx size %u\n",num_rows * sizeof(BASE_TYPE*)); \
	printf("cksize %u\n",num_cols * num_rows * sizeof(BASE_TYPE) + \
		num_rows * sizeof(BASE_TYPE*)); \
	idx = (BASE_TYPE **)memAlloc(num_cols * num_rows * sizeof(BASE_TYPE) + \
		num_rows * sizeof(BASE_TYPE*)); \
	printf("block %u\n",idx); \
	idx[0] = (BASE_TYPE*)(idx + num_rows) - col_base; \
	printf("idx[0] %u\n",idx[0]); \
	for(r = 1;r < num_rows;r++) \
		idx[r] = idx[r-1] + num_cols; \
	return idx - row_base; \
}

#define	MATRIX_DESTRUCTOR(STORAGE,ROUTINE,BASE_TYPE) \
STORAGE	void \
ROUTINE(BASE_TYPE **fake_idx,int row_base) \
{ \
	memFree(fake_idx + row_base); \
}

/*
MATRIX_CONSTRUCTOR(V_LOCAL,new_mat,u8_t)
MATRIX_DESTRUCTOR(V_LOCAL,del_mat,u8_t)
*/

#endif

int main()
{
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
	
	/*
	{
		vec_t x1 = 0, x2 = 3;
		vec_t y1 = 1, y2; // y(0) = 1
		y2 = y1;
		odeint2(&y2,1,x1,x2,1e-4,1e-2,1e-18,derivs);
		printf("was: y(%g) = %g; done: y(%g) = %g\n",x1,y1,x2,y2);
	}
	*/
	/*
	{
		int i,j,count = 0;
		int rb = 10, cb = 15;
		u8_t **m = new_mat(cb,3,rb,4);
		printf("retval %u\n",m);
		printf("data start %u\n",&m[rb][cb]);
		printf("last elem %u\n",&m[rb+3][cb+2]);
		for(i=0;i<4;i++)
			for(j=0;j<3;j++)
				m[rb+i][cb+j] = count++;
		del_mat(m,rb);
	}                  
	*/
	
	/*
	{
		vector_c v(1,2);
	
		v[0] = 0;
		v[1] = 1;
		v[2] = 2;
		v[3] = 3;
	}
	*/
	{
		matrix_c m(1,2,1,2);
	
		m[0][0] = 0;
		m[0][1] = 1;
		m[1][0] = 2;
		m[1][1] = 3;
		
		m[0][4] = 0;
		m[3][0] = 0;
		m[3][4] = 0;
		
	}


	memPrintStats();
	return 0;
}
